from ray_on_golem.exceptions import RayOnGolemError


class RayOnGolemServerError(RayOnGolemError):
    message = None

    def __init__(self, additional_message=None):
        if additional_message:
            self.message += "\n" + additional_message

        super().__init__(self.message)


class NodeNotFound(RayOnGolemServerError):
    message = "Node with given id not found"


class NodesNotFound(RayOnGolemServerError):
    message = "Nodes with given ids not found"


class CreateActivitiesTimeout(RayOnGolemServerError):
    message = "Creating activities timeout reached"


class DestroyActivityError(RayOnGolemServerError):
    message = "Can't destroy activity"


class CheckYagnaStatusError(RayOnGolemServerError):
    message = "Can't check Yagna status"


class ManifestNotFound(RayOnGolemServerError):
    message = "Manifest file not found"


class RegistryRequestError(RayOnGolemServerError):
    message = "Golem Registry request failed"



import asyncio
import hashlib
import logging
import os
from asyncio.subprocess import Process
from collections import deque
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Dict

from aiohttp.web_runner import GracefulExit

from ray_on_golem.exceptions import RayOnGolemError


async def run_subprocess(
    *args, stderr=asyncio.subprocess.DEVNULL, stdout=asyncio.subprocess.DEVNULL
) -> Process:
    process = await asyncio.create_subprocess_exec(
        *args,
        stderr=stderr,
        stdout=stdout,
        # As this process lifetime will be fully managed, we need to disable signal propagation
        # from parent to child process https://stackoverflow.com/a/5446982/1993670
        preexec_fn=os.setpgrp,
    )

    return process


async def run_subprocess_output(*args) -> bytes:
    process = await asyncio.create_subprocess_exec(
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await process.communicate()

    if process.returncode != 0:
        raise RayOnGolemError(
            f"Process exited with code `{process.returncode}`!\nstdout:\n{stdout}\nstderr:\n{stderr}"
        )

    return stdout


def are_dicts_equal(dict1: Dict, dict2: Dict) -> bool:
    for key in dict1.keys():
        if key in dict2:
            if dict1[key] != dict2[key]:
                return False

    return True


def is_running_on_golem_network() -> bool:
    return os.getenv("ON_GOLEM_NETWORK") is not None


def get_default_ssh_key_name(cluster_name: str) -> str:
    return "ray_on_golem_rsa_{}".format(hashlib.md5(cluster_name.encode()).hexdigest()[:10])


def raise_graceful_exit() -> None:
    raise GracefulExit()


def rolloverLogFiles():
    root_logger = logging.getLogger()

    for handler in root_logger.handlers:
        if isinstance(handler, RotatingFileHandler):
            handler.doRollover()


def get_last_lines_from_file(file_path: Path, max_lines: int) -> str:
    with file_path.open() as file:
        return "".join(deque(file, max_lines))



import os
from datetime import timedelta
from pathlib import Path

from yarl import URL

RAY_ON_GOLEM_PATH = Path(os.getenv("RAY_ON_GOLEM_PATH", "ray-on-golem"))
YAGNA_PATH = Path(os.getenv("YAGNA_PATH", "yagna"))
WEBSOCAT_PATH = Path(os.getenv("WEBSOCAT_PATH", "websocat"))
TMP_PATH = Path("/tmp/ray_on_golem")
LOGGING_INFO_PATH = TMP_PATH / "webserver.log"
LOGGING_DEBUG_PATH = TMP_PATH / "webserver_debug.log"
LOGGING_YAGNA_PATH = TMP_PATH / "yagna.log"

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "filters": {
        "add_trace_id": {
            "()": "golem.utils.logging.AddTraceIdFilter",
        },
    },
    "formatters": {
        "compact": {
            "format": "[%(asctime)s] [%(levelname)-7s] [%(name)s] %(message)s",
        },
        "verbose": {
            "format": "[%(asctime)s] [%(levelname)-7s] [%(traceid)s] "
            "[%(name)s:%(lineno)d] %(message)s",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "verbose",
            "filters": ["add_trace_id"],
        },
        "file": {
            "class": "logging.FileHandler",
            "level": "INFO",
            "formatter": "compact",
            "filename": LOGGING_INFO_PATH,
            "mode": "w",
        },
    },
    "root": {
        "level": "INFO",
        "handlers": [
            "console",
            "file",
        ],
    },
    "loggers": {
        "aiohttp": {
            "level": "DEBUG",
            "handlers": ["console"],
            "propagate": False,
        },
        "ray_on_golem": {
            "level": "DEBUG",
        },
        "golem": {
            "level": "INFO",
        },
        "golem.utils.asyncio": {
            "level": "DEBUG",
        },
        "golem.managers.payment": {
            "level": "DEBUG",
        },
        "golem.managers.network": {
            "level": "DEBUG",
        },
        "golem.managers.demand": {
            "level": "DEBUG",
        },
        "golem.managers.proposal": {
            "level": "DEBUG",
        },
        "golem.managers.agreement": {
            "level": "DEBUG",
        },
        "golem.managers.activity": {
            "level": "DEBUG",
        },
        "golem.managers.work": {
            "level": "DEBUG",
        },
    },
}

YAGNA_APPKEY = os.getenv("YAGNA_APPKEY")
YAGNA_APPNAME = os.getenv("YAGNA_APPNAME", "ray-on-golem")
YAGNA_API_URL = URL(os.getenv("YAGNA_API_URL", "http://127.0.0.1:7465"))
YAGNA_START_DEADLINE = timedelta(minutes=2)
YAGNA_CHECK_DEADLINE = timedelta(seconds=2)

RAY_ON_GOLEM_START_DEADLINE = timedelta(minutes=5)
RAY_ON_GOLEM_CHECK_DEADLINE = timedelta(seconds=2)
RAY_ON_GOLEM_SHUTDOWN_DELAY = timedelta(seconds=10)
RAY_ON_GOLEM_SHUTDOWN_DEADLINE = timedelta(seconds=30)

URL_HEALTH_CHECK = "/health_check"
URL_CREATE_CLUSTER = "/create_cluster"
URL_NON_TERMINATED_NODES = "/non_terminated_nodes"
URL_IS_RUNNING = "/is_running"
URL_IS_TERMINATED = "/is_terminated"
URL_NODE_TAGS = "/tags"
URL_INTERNAL_IP = "/internal_ip"
URL_SET_NODE_TAGS = "/set_tags"
URL_CREATE_NODES = "/create_nodes"
URL_TERMINATE_NODE = "/terminate"
URL_GET_SSH_PROXY_COMMAND = "/ssh_proxy_command"
URL_GET_OR_CREATE_DEFAULT_SSH_KEY = "/ger_or_create_default_ssh_key"
URL_SELF_SHUTDOWN = "/self_shutdown"



from typing import Dict

from yarl import URL


def get_manifest(image_url: URL, image_hash: str) -> Dict:
    return {
        "version": "0.1.0",
        "createdAt": "2023-06-26T00:00:00.000000Z",
        "expiresAt": "2100-01-01T00:00:00.000000Z",
        "metadata": {
            "name": "Golem Ray",
            "description": "Golem ray webserver",
            "version": "0.0.1",
        },
        "payload": [
            {
                "urls": [str(image_url)],
                "hash": f"sha3:{image_hash}",
            }
        ],
        "compManifest": {
            "version": "0.1.0",
            "script": {
                "commands": [
                    "run /bin/sh -c *",
                ],
                "match": "regex",
            },
            "net": {
                "inet": {
                    "out": {
                        "protocols": ["https"],
                        "urls": [
                            "https://pypi.dev.golem.network",
                        ],
                    }
                }
            },
        },
    }



import argparse
import logging
import logging.config

from aiohttp import web

from ray_on_golem.server.middlewares import error_middleware, trace_id_middleware
from ray_on_golem.server.services import GolemService, RayService, YagnaService
from ray_on_golem.server.settings import (
    LOGGING_CONFIG,
    RAY_ON_GOLEM_SHUTDOWN_DEADLINE,
    TMP_PATH,
    WEBSOCAT_PATH,
    YAGNA_PATH,
)
from ray_on_golem.server.views import routes

logger = logging.getLogger(__name__)


def parse_sys_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Ray on Golem's webserver.")
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=4578,
        help="port for Ray on Golem's webserver to listen on, default: %(default)s",
    )
    parser.add_argument(
        "--self-shutdown",
        action="store_true",
        help="flag to enable self-shutdown after last node termination, default: %(default)s",
    )
    parser.add_argument("--no-self-shutdown", action="store_false", dest="self_shutdown")
    parser.add_argument(
        "--registry-stats",
        action="store_true",
        help="flag to enable collection of Golem Registry stats about resolved images, default: %(default)s",
    )
    parser.add_argument(
        "--no-registry-stats",
        action="store_false",
        dest="registry_stats",
    )
    parser.set_defaults(self_shutdown=False, registry_stats=True)
    return parser.parse_args()


def prepare_tmp_dir():
    try:
        TMP_PATH.mkdir(parents=True, exist_ok=True)
    except OSError:
        pass


def create_application(port: int, self_shutdown: bool, registry_stats: bool) -> web.Application:
    app = web.Application(
        middlewares=[
            trace_id_middleware,
            error_middleware,
        ]
    )

    app["port"] = port
    app["self_shutdown"] = self_shutdown
    app["registry_stats"] = registry_stats

    app["yagna_service"] = YagnaService(
        yagna_path=YAGNA_PATH,
    )

    app["golem_service"] = GolemService(
        websocat_path=WEBSOCAT_PATH,
        registry_stats=app["registry_stats"],
    )

    app["ray_service"] = RayService(
        ray_on_golem_port=port,
        golem_service=app["golem_service"],
        tmp_path=TMP_PATH,
    )

    app.add_routes(routes)
    app.cleanup_ctx.append(yagna_service_ctx)
    app.cleanup_ctx.append(golem_service_ctx)
    app.cleanup_ctx.append(ray_service_ctx)
    app.on_startup.append(startup_print)
    app.on_shutdown.append(shutdown_print)

    return app


async def startup_print(app: web.Application) -> None:
    logger.info("Starting server done, listening on port {}".format(app["port"]))


async def shutdown_print(app: web.Application) -> None:
    logger.info("Stopping server gracefully, forcing after `%s`...", RAY_ON_GOLEM_SHUTDOWN_DEADLINE)


async def yagna_service_ctx(app: web.Application) -> None:
    yagna_service: YagnaService = app["yagna_service"]

    await yagna_service.init()

    yield

    await yagna_service.shutdown()


async def golem_service_ctx(app: web.Application) -> None:
    golem_service: GolemService = app["golem_service"]
    yagna_service: YagnaService = app["yagna_service"]

    await golem_service.init(yagna_appkey=yagna_service.yagna_appkey)

    yield

    await golem_service.shutdown()


async def ray_service_ctx(app: web.Application) -> None:
    ray_service: RayService = app["ray_service"]

    yield

    await ray_service.shutdown()


def main():
    prepare_tmp_dir()
    args = parse_sys_args()

    logging.config.dictConfig(LOGGING_CONFIG)

    app = create_application(args.port, args.self_shutdown, args.registry_stats)

    logger.info(
        "Starting server... {}".format(", ".join(f"{k}={v}" for k, v in args.__dict__.items()))
    )

    try:
        web.run_app(
            app,
            port=app["port"],
            print=None,
            shutdown_timeout=RAY_ON_GOLEM_SHUTDOWN_DEADLINE.total_seconds(),
        )
    except Exception:
        logger.info("Server unexpectedly died, bye!")
    else:
        logger.info("Stopping server done, bye!")


if __name__ == "__main__":
    main()



from pathlib import Path


def get_ssh_command(
    ip: str, ssh_proxy_command: str, ssh_user: str, ssh_private_key_path: Path
) -> str:
    return (
        "ssh "
        "-o StrictHostKeyChecking=no "
        "-o UserKnownHostsFile=/dev/null "
        f'-o "ProxyCommand={ssh_proxy_command}" '
        f"-i {ssh_private_key_path} "
        f"{ssh_user}@{ip}"
    )



from enum import Enum
from typing import Any, Dict, List, Optional

from golem.resources import Activity
from pydantic import BaseModel, root_validator

NodeId = str
Tags = Dict[str, str]


class NodeState(Enum):
    pending = "pending"
    running = "running"
    stopping = "stopping"


class ShutdownState(Enum):
    NOT_ENABLED = "not_enabled"
    CLUSTER_NOT_EMPTY = "cluster_not_empty"
    WILL_SHUTDOWN = "will_shutdown"


class Node(BaseModel):
    node_id: NodeId
    state: NodeState
    tags: Tags
    internal_ip: str
    ssh_proxy_command: str
    activity: Activity

    class Config:
        arbitrary_types_allowed = True


class SingleNodeRequestData(BaseModel):
    node_id: NodeId


class DemandConfigData(BaseModel):
    image_hash: Optional[str] = None
    image_tag: Optional[str] = None
    capabilities: List[str]
    min_mem_gib: float
    min_cpu_threads: int
    min_storage_gib: float


class CostManagementData(BaseModel):
    average_cpu_load: Optional[float] = None
    average_duration_minutes: Optional[float] = None

    max_average_usage_cost: Optional[float] = None
    max_initial_price: Optional[float] = None
    max_cpu_sec_price: Optional[float] = None
    max_duration_sec_price: Optional[float] = None

    @root_validator
    def check_average_fields(cls, values):
        average_cpu_load = values.get("average_cpu_load")
        average_duration_minutes = values.get("average_duration_minutes")
        max_average_usage_cost = values.get("max_average_usage_cost")

        if average_cpu_load is None != average_duration_minutes is None:
            raise ValueError(
                "Both `average_cpu_load` and `average_duration_minutes` parameter should be defined together!"
            )

        if max_average_usage_cost is not None and (
            average_cpu_load is None or average_duration_minutes is None
        ):
            raise ValueError(
                "Parameter `max_average_usage_cost` requires `average_cpu_load` and `average_duration_minutes`"
                " parameters to be defined!"
            )

        return values

    def is_average_usage_cost_enabled(self):
        return self.average_cpu_load is not None and self.average_duration_minutes is not None


class NodeConfigData(BaseModel):
    demand: DemandConfigData
    cost_management: Optional[CostManagementData] = None


class ProviderConfigData(BaseModel):
    network: str
    budget: int
    node_config: NodeConfigData
    ssh_private_key: str
    ssh_user: str


class CreateClusterRequestData(ProviderConfigData):
    pass


class NonTerminatedNodesRequestData(BaseModel):
    tags: Tags


class NonTerminatedNodesResponseData(BaseModel):
    nodes_ids: List[NodeId]


class CreateNodesRequestData(BaseModel):
    node_config: Dict[str, Any]
    count: int
    tags: Tags


class CreateNodesResponseData(BaseModel):
    created_nodes: Dict[NodeId, Dict]


class SetNodeTagsRequestData(BaseModel):
    node_id: NodeId
    tags: Tags


class TerminateNodeResponseData(BaseModel):
    terminated_nodes: Dict[NodeId, Dict]


class IsRunningResponseData(BaseModel):
    is_running: bool


class IsTerminatedResponseData(BaseModel):
    is_terminated: bool


class GetNodeTagsResponseData(BaseModel):
    tags: Tags


class GetNodeIpAddressResponseData(BaseModel):
    ip_address: str


class EmptyResponseData(BaseModel):
    pass


class GetSshProxyCommandResponseData(BaseModel):
    ssh_proxy_command: str


class GetOrCreateDefaultSshKeyRequestData(BaseModel):
    cluster_name: str


class GetOrCreateDefaultSshKeyResponseData(BaseModel):
    ssh_key_base64: str


class SelfShutdownRequestData(BaseModel):
    pass


class SelfShutdownResponseData(BaseModel):
    shutdown_state: ShutdownState



import uuid

from aiohttp import web
from golem.utils.logging import trace_id_var

from ray_on_golem.server.exceptions import RayOnGolemServerError


@web.middleware
async def trace_id_middleware(request, handler):
    trace_id_var.set(f"webserver-request-{str(uuid.uuid4())[:5]}")

    return await handler(request)


@web.middleware
async def error_middleware(request, handler):
    try:
        response = await handler(request)
    except web.HTTPException as e:
        message = e.reason
        status_code = e.status_code
    except RayOnGolemServerError as e:
        message = e.message
        status_code = 400
    else:
        if response.status != 404:
            return response

        message = "not found"
        status_code = 404

    return web.json_response({"error": message}, status=status_code)



from ray_on_golem.server.services.golem import GolemService, get_manifest
from ray_on_golem.server.services.ray import RayService
from ray_on_golem.server.services.yagna import YagnaService

__all__ = (
    "GolemService",
    "RayService",
    "YagnaService",
    "get_manifest",
)



import logging
import subprocess
from copy import deepcopy
from datetime import datetime
from functools import lru_cache
from types import ModuleType
from typing import Any, Dict, List, Optional

from ray.autoscaler._private.cli_logger import cli_logger
from ray.autoscaler._private.event_system import CreateClusterEvent, global_event_system
from ray.autoscaler.command_runner import CommandRunnerInterface
from ray.autoscaler.node_provider import NodeProvider

from ray_on_golem.client.client import RayOnGolemClient
from ray_on_golem.provider.ssh_command_runner import SSHCommandRunner
from ray_on_golem.server.models import NodeConfigData, NodeId, ShutdownState
from ray_on_golem.server.run import prepare_tmp_dir
from ray_on_golem.server.settings import (
    LOGGING_DEBUG_PATH,
    RAY_ON_GOLEM_CHECK_DEADLINE,
    RAY_ON_GOLEM_PATH,
    RAY_ON_GOLEM_START_DEADLINE,
    TMP_PATH,
)
from ray_on_golem.utils import (
    get_default_ssh_key_name,
    get_last_lines_from_file,
    is_running_on_golem_network,
)

logger = logging.getLogger(__name__)
WEBSERVER_LOG_GROUP = "Ray On Golem webserver"

WEBSERVER_OUTFILE = TMP_PATH / "webserver.out"
WEBSERVER_ERRFILE = TMP_PATH / "webserver.err"


class GolemNodeProvider(NodeProvider):
    def __init__(self, provider_config: dict, cluster_name: str):
        super().__init__(provider_config, cluster_name)

        provider_parameters = provider_config["parameters"]

        self._ray_on_golem_client = self._get_ray_on_golem_client_instance(
            provider_parameters["webserver_port"],
            provider_parameters["enable_registry_stats"],
        )
        self._ray_on_golem_client.create_cluster(
            network=provider_parameters["network"],
            budget=provider_parameters["budget"],
            node_config=NodeConfigData(**provider_parameters["node_config"]),
            ssh_private_key=provider_parameters["_ssh_private_key"],
            ssh_user=provider_parameters["_ssh_user"],
        )

    @classmethod
    def bootstrap_config(cls, cluster_config: Dict[str, Any]) -> Dict[str, Any]:
        config = deepcopy(cluster_config)

        provider_parameters: Dict = config["provider"]["parameters"]
        provider_parameters.setdefault("webserver_port", 4578)
        provider_parameters.setdefault("enable_registry_stats", True)
        provider_parameters.setdefault("network", "goerli")
        provider_parameters.setdefault("budget", 1)

        ray_on_golem_client = cls._get_ray_on_golem_client_instance(
            provider_parameters["webserver_port"],
            provider_parameters["enable_registry_stats"],
        )

        auth: Dict = config["auth"]
        auth.setdefault("ssh_user", "root")

        if "ssh_private_key" not in auth:
            ssh_key_path = TMP_PATH / get_default_ssh_key_name(config["cluster_name"])
            auth["ssh_private_key"] = str(ssh_key_path)

            if not ssh_key_path.exists():
                ssh_key_base64 = ray_on_golem_client.get_or_create_default_ssh_key(
                    config["cluster_name"]
                )

                # FIXME: mitigate double file writing on local machine as get_or_create_default_ssh_key creates the file
                ssh_key_path.parent.mkdir(parents=True, exist_ok=True)
                with ssh_key_path.open("w") as f:
                    f.write(ssh_key_base64)

        # copy ssh details to provider namespace for cluster creation in __init__
        provider_parameters["_ssh_private_key"] = auth["ssh_private_key"]
        provider_parameters["_ssh_user"] = auth["ssh_user"]

        global_event_system.execute_callback(
            CreateClusterEvent.ssh_keypair_downloaded,
            {"ssh_key_path": auth["ssh_private_key"]},
        )

        return config

    @classmethod
    @lru_cache()
    def _get_ray_on_golem_client_instance(cls, webserver_port: int, enable_registry_stats: bool):
        ray_on_golem_client = RayOnGolemClient(webserver_port)

        if not is_running_on_golem_network():
            cls._start_webserver(ray_on_golem_client, webserver_port, enable_registry_stats)

        return ray_on_golem_client

    def get_command_runner(
        self,
        log_prefix: str,
        node_id: str,
        auth_config: Dict[str, Any],
        cluster_name: str,
        process_runner: ModuleType,
        use_internal_ip: bool,
        docker_config: Optional[Dict[str, Any]] = None,
    ) -> CommandRunnerInterface:
        common_args = {
            "log_prefix": log_prefix,
            "node_id": node_id,
            "provider": self,
            "auth_config": auth_config,
            "cluster_name": cluster_name,
            "process_runner": process_runner,
            "use_internal_ip": use_internal_ip,
        }

        if "ssh_proxy_command" not in auth_config and not is_running_on_golem_network():
            auth_config["ssh_proxy_command"] = self._ray_on_golem_client.get_ssh_proxy_command(
                node_id
            )

        return SSHCommandRunner(**common_args)

    def create_node(
        self,
        node_config: Dict[str, Any],
        tags: Dict[str, str],
        count: int,
    ) -> Dict[NodeId, Dict]:
        return self._ray_on_golem_client.create_nodes(
            node_config=node_config,
            count=count,
            tags=tags,
        )

    def terminate_node(self, node_id: NodeId) -> Dict[NodeId, Dict]:
        terminated_nodes = self._ray_on_golem_client.terminate_node(node_id)

        self._stop_webserver(self._ray_on_golem_client)

        return terminated_nodes

    def non_terminated_nodes(self, tag_filters) -> List[NodeId]:
        return self._ray_on_golem_client.non_terminated_nodes(tag_filters)

    def is_running(self, node_id: NodeId) -> bool:
        return self._ray_on_golem_client.is_running(node_id)

    def is_terminated(self, node_id: NodeId) -> bool:
        return self._ray_on_golem_client.is_terminated(node_id)

    def node_tags(self, node_id: NodeId) -> Dict:
        return self._ray_on_golem_client.get_node_tags(node_id)

    def internal_ip(self, node_id: NodeId) -> str:
        return self._ray_on_golem_client.get_node_internal_ip(node_id)

    def external_ip(self, node_id: NodeId) -> str:
        return self._ray_on_golem_client.get_node_internal_ip(node_id)

    def set_node_tags(self, node_id: NodeId, tags: Dict) -> None:
        self._ray_on_golem_client.set_node_tags(node_id, tags)

    @staticmethod
    def _start_webserver(
        ray_on_golem_client: RayOnGolemClient,
        port: int,
        registry_stats: bool,
    ) -> None:
        with cli_logger.group(WEBSERVER_LOG_GROUP):
            if ray_on_golem_client.is_webserver_running():
                cli_logger.print("Not starting webserver, as it's already running")
                return

            cli_logger.print(
                "Starting webserver with deadline up to `{}`...", RAY_ON_GOLEM_START_DEADLINE
            )
            args = [
                RAY_ON_GOLEM_PATH,
                "-p",
                str(port),
                "--registry-stats" if registry_stats else "--no-registry-stats",
                "--self-shutdown",
            ]

            cli_logger.verbose("Webserver command: `{}`", " ".join([str(a) for a in args]))

            prepare_tmp_dir()
            log_file = LOGGING_DEBUG_PATH.open("w")
            proc = subprocess.Popen(
                args,
                stdout=log_file,
                stderr=log_file,
                start_new_session=True,
            )

            start_deadline = datetime.now() + RAY_ON_GOLEM_START_DEADLINE
            check_seconds = int(RAY_ON_GOLEM_CHECK_DEADLINE.total_seconds())
            while datetime.now() < start_deadline:
                try:
                    proc.communicate(timeout=check_seconds)
                except subprocess.TimeoutExpired:
                    if ray_on_golem_client.is_webserver_running():
                        cli_logger.print("Starting webserver done")
                        return
                else:
                    cli_logger.abort(
                        "Starting webserver failed!\nShowing last 50 lines from `{}`:\n{}",
                        LOGGING_DEBUG_PATH,
                        get_last_lines_from_file(LOGGING_DEBUG_PATH, 50),
                    )

                cli_logger.print(
                    "Webserver is not yet running, waiting additional `{}` seconds...",
                    check_seconds,
                )

            cli_logger.abort(
                "Starting webserver failed! Deadline of `{}` reached.\nShowing last 50 lines from `{}`:\n{}",
                RAY_ON_GOLEM_START_DEADLINE,
                LOGGING_DEBUG_PATH,
                get_last_lines_from_file(LOGGING_DEBUG_PATH, 50),
            )

    @staticmethod
    def _stop_webserver(ray_on_golem_client: RayOnGolemClient) -> None:
        with cli_logger.group(WEBSERVER_LOG_GROUP):
            if not ray_on_golem_client.is_webserver_running():
                cli_logger.print("Not stopping webserver, as it's already not running")
                return

            cli_logger.print("Requesting webserver shutdown...")

            shutdown_state = ray_on_golem_client.shutdown_webserver()

            if shutdown_state == ShutdownState.NOT_ENABLED:
                cli_logger.print("Not stopping webserver, as it was started externally")
                return
            elif shutdown_state == ShutdownState.CLUSTER_NOT_EMPTY:
                cli_logger.print("Not stopping webserver, as the cluster is not empty")
                return

            cli_logger.print("Requesting webserver shutdown done, will stop soon")



from typing import Any, Dict, List, Type, TypeVar

import requests
from pydantic import BaseModel, ValidationError
from yarl import URL

from ray_on_golem.client.exceptions import RayOnGolemClientError, RayOnGolemClientValidationError
from ray_on_golem.server import models, settings
from ray_on_golem.server.models import NodeConfigData

TResponseModel = TypeVar("TResponseModel")


class RayOnGolemClient:
    def __init__(self, port: int) -> None:
        self._base_url = URL("http://127.0.0.1").with_port(port)

        self._session = requests.Session()

    def create_cluster(
        self,
        network: str,
        budget: int,
        node_config: NodeConfigData,
        ssh_private_key: str,
        ssh_user: str,
    ) -> None:
        self._make_request(
            url=settings.URL_CREATE_CLUSTER,
            request_data=models.CreateClusterRequestData(
                network=network,
                budget=budget,
                node_config=node_config,
                ssh_private_key=ssh_private_key,
                ssh_user=ssh_user,
            ),
            response_model=models.EmptyResponseData,
            error_message="Couldn't create cluster",
        )

    def create_nodes(
        self, node_config: Dict[str, Any], count: int, tags: models.Tags
    ) -> Dict[models.NodeId, Dict]:
        response = self._make_request(
            url=settings.URL_CREATE_NODES,
            request_data=models.CreateNodesRequestData(
                node_config=node_config,
                count=count,
                tags=tags,
            ),
            response_model=models.CreateNodesResponseData,
            error_message="Couldn't create node",
        )

        return response.created_nodes

    def terminate_node(self, node_id: models.NodeId) -> Dict[models.NodeId, Dict]:
        response = self._make_request(
            url=settings.URL_TERMINATE_NODE,
            request_data=models.SingleNodeRequestData(
                node_id=node_id,
            ),
            response_model=models.TerminateNodeResponseData,
            error_message="Couldn't terminate node",
        )

        return response.terminated_nodes

    def non_terminated_nodes(self, tag_filters: models.Tags) -> List[models.NodeId]:
        response = self._make_request(
            url=settings.URL_NON_TERMINATED_NODES,
            request_data=models.NonTerminatedNodesRequestData(
                tags=tag_filters,
            ),
            response_model=models.NonTerminatedNodesResponseData,
            error_message="Couldn't get non terminated nodes",
        )

        return response.nodes_ids

    def is_running(self, node_id: models.NodeId) -> bool:
        response = self._make_request(
            url=settings.URL_IS_RUNNING,
            request_data=models.SingleNodeRequestData(
                node_id=node_id,
            ),
            response_model=models.IsRunningResponseData,
            error_message="Couldn't check if node is running",
        )

        return response.is_running

    def is_terminated(self, node_id: models.NodeId) -> bool:
        response = self._make_request(
            url=settings.URL_IS_TERMINATED,
            request_data=models.SingleNodeRequestData(
                node_id=node_id,
            ),
            response_model=models.IsTerminatedResponseData,
            error_message="Couldn't check if node is terminated",
        )

        return response.is_terminated

    def get_node_tags(self, node_id: models.NodeId) -> models.Tags:
        response = self._make_request(
            url=settings.URL_NODE_TAGS,
            request_data=models.SingleNodeRequestData(
                node_id=node_id,
            ),
            response_model=models.GetNodeTagsResponseData,
            error_message="Couldn't get node tags",
        )

        return response.tags

    def get_node_internal_ip(self, node_id: models.NodeId) -> str:
        response = self._make_request(
            url=settings.URL_INTERNAL_IP,
            request_data=models.SingleNodeRequestData(
                node_id=node_id,
            ),
            response_model=models.GetNodeIpAddressResponseData,
            error_message="Couldn't get node internal_ip",
        )

        return str(response.ip_address)

    def set_node_tags(self, node_id: models.NodeId, tags: models.Tags) -> None:
        self._make_request(
            url=settings.URL_SET_NODE_TAGS,
            request_data=models.SetNodeTagsRequestData(
                node_id=node_id,
                tags=tags,
            ),
            response_model=models.EmptyResponseData,
            error_message="Couldn't set tags for node",
        )

    def get_ssh_proxy_command(self, node_id: str) -> str:
        response = self._make_request(
            url=settings.URL_GET_SSH_PROXY_COMMAND,
            request_data=models.SingleNodeRequestData(
                node_id=node_id,
            ),
            response_model=models.GetSshProxyCommandResponseData,
            error_message="Couldn't get ssh proxy command",
        )

        return response.ssh_proxy_command

    def get_or_create_default_ssh_key(self, cluster_name: str) -> str:
        response = self._make_request(
            url=settings.URL_GET_OR_CREATE_DEFAULT_SSH_KEY,
            request_data=models.GetOrCreateDefaultSshKeyRequestData(
                cluster_name=cluster_name,
            ),
            response_model=models.GetOrCreateDefaultSshKeyResponseData,
            error_message="Couldn't get or create default ssh key",
        )

        return response.ssh_key_base64

    def shutdown_webserver(self) -> models.ShutdownState:
        response = self._make_request(
            url=settings.URL_SELF_SHUTDOWN,
            request_data=models.SelfShutdownRequestData(),
            response_model=models.SelfShutdownResponseData,
            error_message="Couldn't send a self-shutdown request",
        )

        return response.shutdown_state

    def is_webserver_running(self) -> bool:
        try:
            response = requests.get(
                str(self._base_url / settings.URL_HEALTH_CHECK.lstrip("/")),
                timeout=settings.RAY_ON_GOLEM_CHECK_DEADLINE.total_seconds(),
            )
        except requests.ConnectionError:
            return False
        else:
            return response.status_code == 200 and response.text == "ok"

    def _make_request(
        self,
        *,
        url: str,
        request_data: BaseModel,
        response_model: Type[TResponseModel],
        error_message: str,
    ) -> TResponseModel:
        response = self._session.post(
            str(self._base_url / url.lstrip("/")), data=request_data.json()
        )

        if response.status_code != 200:
            raise RayOnGolemClientError(f"{error_message}: {response.text}")

        try:
            return response_model.parse_raw(response.text)
        except ValidationError as e:
            raise RayOnGolemClientValidationError(
                "Couldn't validate response data",
            ) from e



import asyncio
import logging

from aiohttp import web

from ray_on_golem.server import models, settings
from ray_on_golem.server.models import ShutdownState
from ray_on_golem.server.services import RayService
from ray_on_golem.utils import raise_graceful_exit

logger = logging.getLogger(__name__)

routes = web.RouteTableDef()


# FIXME: This route should be a default root URL with basic server info instead of
#  custom URL with custom payload
@routes.get(settings.URL_HEALTH_CHECK)
async def health_check(request: web.Request) -> web.Response:
    return web.Response(text="ok")


@routes.post(settings.URL_CREATE_CLUSTER)
async def create_cluster(request: web.Request) -> web.Response:
    ray_service: RayService = request.app["ray_service"]

    request_data = models.CreateClusterRequestData.parse_raw(await request.text())

    await ray_service.create_cluster(provider_config=request_data)

    response_data = models.EmptyResponseData()

    return web.Response(text=response_data.json())


@routes.post(settings.URL_NON_TERMINATED_NODES)
async def non_terminated_nodes_ids(request: web.Request) -> web.Response:
    ray_service: RayService = request.app["ray_service"]

    request_data = models.NonTerminatedNodesRequestData.parse_raw(await request.text())

    nodes_ids = await ray_service.get_non_terminated_nodes_ids(tags_to_match=request_data.tags)

    response_data = models.NonTerminatedNodesResponseData(nodes_ids=nodes_ids)

    return web.Response(text=response_data.json())


@routes.post(settings.URL_IS_RUNNING)
async def is_node_running(request: web.Request) -> web.Response:
    ray_service: RayService = request.app["ray_service"]

    request_data = models.SingleNodeRequestData.parse_raw(await request.text())

    is_running = await ray_service.is_node_running(request_data.node_id)

    response_data = models.IsRunningResponseData(is_running=is_running)

    return web.Response(text=response_data.json())


@routes.post(settings.URL_IS_TERMINATED)
async def is_node_terminated(request: web.Request) -> web.Response:
    ray_service: RayService = request.app["ray_service"]

    request_data = models.SingleNodeRequestData.parse_raw(await request.text())

    is_terminated = await ray_service.is_node_terminated(request_data.node_id)

    response_data = models.IsTerminatedResponseData(is_terminated=is_terminated)

    return web.Response(text=response_data.json())


@routes.post(settings.URL_NODE_TAGS)
async def get_node_tags(request: web.Request) -> web.Response:
    ray_service: RayService = request.app["ray_service"]

    request_data = models.SingleNodeRequestData.parse_raw(await request.text())

    node_tags = await ray_service.get_node_tags(request_data.node_id)

    response_data = models.GetNodeTagsResponseData(tags=node_tags)

    return web.Response(text=response_data.json())


@routes.post(settings.URL_INTERNAL_IP)
async def get_node_internal_ip(request: web.Request) -> web.Response:
    ray_service: RayService = request.app["ray_service"]

    request_data = models.SingleNodeRequestData.parse_raw(await request.text())

    ip_address = await ray_service.get_node_internal_ip(request_data.node_id)

    response_data = models.GetNodeIpAddressResponseData(ip_address=ip_address)

    return web.Response(text=response_data.json())


@routes.post(settings.URL_SET_NODE_TAGS)
async def set_node_tags(request: web.Request) -> web.Response:
    ray_service: RayService = request.app["ray_service"]

    request_data = models.SetNodeTagsRequestData.parse_raw(await request.text())

    await ray_service.set_node_tags(
        node_id=request_data.node_id,
        tags=request_data.tags,
    )

    response_data = models.EmptyResponseData()

    return web.Response(text=response_data.json())


@routes.post(settings.URL_CREATE_NODES)
async def create_nodes(request: web.Request) -> web.Response:
    ray_service: RayService = request.app["ray_service"]

    request_data = models.CreateNodesRequestData.parse_raw(await request.text())

    created_nodes = await ray_service.create_nodes(
        request_data.node_config, request_data.count, request_data.tags
    )

    response_data = models.CreateNodesResponseData(created_nodes=created_nodes)

    return web.Response(text=response_data.json())


@routes.post(settings.URL_TERMINATE_NODE)
async def terminate_node(request: web.Request) -> web.Response:
    ray_service: RayService = request.app["ray_service"]

    request_data = models.SingleNodeRequestData.parse_raw(await request.text())

    terminated_nodes = await ray_service.terminate_node(request_data.node_id)

    response_data = models.TerminateNodeResponseData(terminated_nodes=terminated_nodes)

    return web.Response(text=response_data.json())


@routes.post(settings.URL_GET_SSH_PROXY_COMMAND)
async def get_ssh_proxy_command(request):
    ray_service: RayService = request.app["ray_service"]

    request_data = models.SingleNodeRequestData.parse_raw(await request.text())

    ssh_proxy_command = await ray_service.get_ssh_proxy_command(node_id=request_data.node_id)

    response_data = models.GetSshProxyCommandResponseData(ssh_proxy_command=ssh_proxy_command)

    return web.Response(text=response_data.json())


@routes.post(settings.URL_GET_OR_CREATE_DEFAULT_SSH_KEY)
async def get_or_create_ssh_key(request):
    ray_service: RayService = request.app["ray_service"]

    request_data = models.GetOrCreateDefaultSshKeyRequestData.parse_raw(await request.text())

    ssh_key_base64 = await ray_service.get_or_create_default_ssh_key(
        cluster_name=request_data.cluster_name
    )

    response_data = models.GetOrCreateDefaultSshKeyResponseData(ssh_key_base64=ssh_key_base64)

    return web.Response(text=response_data.json())


@routes.post(settings.URL_SELF_SHUTDOWN)
async def self_shutdown(request):
    ray_service: RayService = request.app["ray_service"]

    models.SelfShutdownRequestData.parse_raw(await request.text())

    if not request.app["self_shutdown"]:
        shutdown_state = ShutdownState.NOT_ENABLED
    elif await ray_service.get_non_terminated_nodes_ids():
        shutdown_state = ShutdownState.CLUSTER_NOT_EMPTY
    else:
        shutdown_state = ShutdownState.WILL_SHUTDOWN

    if shutdown_state == ShutdownState.WILL_SHUTDOWN:
        shutdown_seconds = int(settings.RAY_ON_GOLEM_SHUTDOWN_DELAY.total_seconds())
        logger.info(f"Received a self-shutdown request, exiting in {shutdown_seconds} seconds...")
        loop = asyncio.get_event_loop()
        loop.call_later(shutdown_seconds, raise_graceful_exit)

    response_data = models.SelfShutdownResponseData(shutdown_state=shutdown_state)

    return web.Response(text=response_data.json())



import asyncio
import json
import logging
from asyncio.subprocess import Process
from datetime import datetime
from pathlib import Path
from typing import Optional

import aiohttp

from ray_on_golem.exceptions import RayOnGolemError
from ray_on_golem.server.settings import (
    LOGGING_YAGNA_PATH,
    RAY_ON_GOLEM_CHECK_DEADLINE,
    YAGNA_API_URL,
    YAGNA_APPKEY,
    YAGNA_APPNAME,
    YAGNA_CHECK_DEADLINE,
    YAGNA_START_DEADLINE,
)
from ray_on_golem.utils import get_last_lines_from_file, run_subprocess, run_subprocess_output

logger = logging.getLogger(__name__)


class YagnaServiceError(RayOnGolemError):
    pass


class YagnaService:
    def __init__(self, yagna_path: Path):
        self._yagna_path = yagna_path

        self.yagna_appkey: Optional[str] = None
        self._yagna_process: Optional[Process] = None
        self._yagna_early_exit_task: Optional[asyncio.Task] = None

    async def init(self) -> None:
        logger.info("Starting YagnaService...")

        if await self._check_if_yagna_is_running():
            logger.info("Not starting Yagna, as it was started externally")
        else:
            await self._run_yagna()
            await self._run_yagna_payment_fund()

        self.yagna_appkey = await self._get_or_create_yagna_appkey()

        logger.info("Starting YagnaService done")

    async def shutdown(self) -> None:
        logger.info("Stopping YagnaService...")

        await self._stop_yagna_service()

        logger.info("Stopping YagnaService done")

    async def _wait_for_yagna_api(self) -> bool:
        logger.debug("Waiting for Yagna Api...")

        check_seconds = int(RAY_ON_GOLEM_CHECK_DEADLINE.total_seconds())
        for _ in range(25):
            if await self._check_if_yagna_is_running():
                logger.debug("Waiting for Yagna Api done")
                return True

            logger.debug(
                "Yagna Api not yet ready, waiting additional `%s` seconds...", check_seconds
            )
            await asyncio.sleep(check_seconds)

        logger.debug("Waiting for Yagna Api failed with timeout!")

        return False

    async def _check_if_yagna_is_running(self) -> bool:
        try:
            async with aiohttp.ClientSession() as client:
                async with client.get(YAGNA_API_URL):
                    return True
        except aiohttp.ClientError:
            return False

    async def _run_yagna(self) -> None:
        logger.info("Starting Yagna...")

        log_file = LOGGING_YAGNA_PATH.open("w")
        self._yagna_process = await run_subprocess(
            self._yagna_path, "service", "run", stderr=log_file, stdout=log_file
        )

        start_deadline = datetime.now() + YAGNA_START_DEADLINE
        check_seconds = int(YAGNA_CHECK_DEADLINE.total_seconds())
        while datetime.now() < start_deadline:
            try:
                await asyncio.wait_for(self._yagna_process.communicate(), timeout=check_seconds)
            except asyncio.TimeoutError:
                if await self._check_if_yagna_is_running():
                    self._yagna_early_exit_task = asyncio.create_task(self._on_yagna_early_exit())
                    logger.info("Starting Yagna done")
                    return
            else:
                logger.error(
                    "Starting Yagna failed!\nShowing last 50 lines from `%s`:\n%s",
                    LOGGING_YAGNA_PATH,
                    get_last_lines_from_file(LOGGING_YAGNA_PATH, 50),
                )
                raise YagnaServiceError("Starting Yagna failed!")

            logger.info(
                "Yagna is not yet running, waiting additional `%s` seconds...",
                check_seconds,
            )

        logger.error(
            "Starting Yagna failed! Deadline of `%s` reached.\nShowing last 50 lines from `%s`:\n%s",
            YAGNA_START_DEADLINE,
            LOGGING_YAGNA_PATH,
            get_last_lines_from_file(LOGGING_YAGNA_PATH, 50),
        )
        raise YagnaServiceError("Starting Yagna failed!")

    async def _on_yagna_early_exit(self) -> None:
        await self._yagna_process.communicate()

        logger.warning(f"Yagna exited prematurely!")

    async def _stop_yagna_service(self) -> None:
        if self._yagna_process is None:
            logger.info("Not stopping Yagna, as it was started externally")
            return

        if self._yagna_process.returncode is not None:
            logger.info("Not stopping Yagna, as it's not running")
            return

        logger.info("Stopping Yagna...")

        self._yagna_early_exit_task.cancel()
        try:
            await self._yagna_early_exit_task
        except asyncio.CancelledError:
            pass

        self._yagna_process.terminate()
        await self._yagna_process.wait()

        logger.info("Stopping Yagna done")

    async def _run_yagna_payment_fund(self) -> None:
        await run_subprocess_output(self._yagna_path, "payment", "fund")

    async def _get_or_create_yagna_appkey(self) -> str:
        if YAGNA_APPKEY:
            return YAGNA_APPKEY

        output = await run_subprocess_output(self._yagna_path, "app-key", "list", "--json")

        yagna_app = next((app for app in json.loads(output) if app["name"] == YAGNA_APPNAME), None)

        if yagna_app is not None:
            return yagna_app["key"]

        output = await run_subprocess_output(
            self._yagna_path,
            "app-key",
            "create",
            YAGNA_APPNAME,
            "--json",
        )

        return json.loads(output)



import asyncio
import logging
from asyncio.subprocess import Process
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional

from ray.autoscaler.tags import NODE_KIND_HEAD, TAG_RAY_NODE_KIND

from ray_on_golem.exceptions import RayOnGolemError
from ray_on_golem.server.exceptions import NodeNotFound
from ray_on_golem.server.models import Node, NodeId, NodeState, ProviderConfigData, Tags
from ray_on_golem.server.services.golem import GolemService
from ray_on_golem.server.services.utils import get_ssh_command
from ray_on_golem.utils import (
    are_dicts_equal,
    get_default_ssh_key_name,
    run_subprocess,
    run_subprocess_output,
)

logger = logging.getLogger(__name__)


class RayServiceError(RayOnGolemError):
    pass


class RayService:
    def __init__(self, ray_on_golem_port: int, golem_service: GolemService, tmp_path: Path):
        self._ray_on_golem_port = ray_on_golem_port
        self._golem_service = golem_service
        self._tmp_path = tmp_path

        self._provider_config: Optional[ProviderConfigData] = None

        self._nodes: Dict[NodeId, Node] = {}
        self._nodes_lock = asyncio.Lock()
        self._nodes_id_counter = 0

        self._head_node_to_webserver_tunnel_process: Optional[Process] = None
        self._head_node_to_webserver_tunnel_early_exit_task: Optional[asyncio.Task] = None

        self._ssh_private_key_path: Optional[Path] = None
        self._ssh_public_key_path: Optional[Path] = None
        self._ssh_public_key: Optional[str] = None
        self._ssh_user: Optional[str] = None

    async def shutdown(self) -> None:
        logger.info("Stopping RayService...")

        await self._stop_head_node_to_webserver_tunnel()
        await self._destroy_nodes()

        logger.info("Stopping RayService done")

    async def create_cluster(self, provider_config: ProviderConfigData) -> None:
        self._provider_config = provider_config

        self._ssh_private_key_path = Path(provider_config.ssh_private_key)
        self._ssh_public_key_path = self._ssh_private_key_path.with_suffix(".pub")
        self._ssh_user = provider_config.ssh_user

        with self._ssh_public_key_path.open() as f:
            self._ssh_public_key = f.readline().strip()

    async def _destroy_nodes(self) -> None:
        async with self._nodes_lock:
            if not self._nodes:
                logger.info("Not destroying nodes, as no nodes are running")
                return

            logger.info(f"Destroying {len(self._nodes)} nodes...")

            for node in self._nodes.values():
                await node.activity.destroy()

            logger.info(f"Destroying {len(self._nodes)} nodes done")

            self._nodes.clear()

    async def create_nodes(
        self, node_config: Dict[str, Any], count: int, tags: Tags
    ) -> Dict[NodeId, Dict]:
        # TODO: handle pending state
        # TODO: Use node_config from yaml.available_node_types not, from yaml.provider

        if not self._provider_config:
            raise RayServiceError("Node creation is available only after cluster bootstrap!")

        logger.info(f"Creating {count} nodes...")

        created_nodes = {}
        async for activity, ip, ssh_proxy_command in self._golem_service.create_activities(
            node_config=self._provider_config.node_config,
            count=count,
            ssh_public_key=self._ssh_public_key,
            ssh_user=self._ssh_user,
            ssh_private_key_path=self._ssh_private_key_path,
            budget=self._provider_config.budget,
            network=self._provider_config.network,
        ):
            self._print_ssh_command(
                ip, ssh_proxy_command, self._ssh_user, self._ssh_private_key_path
            )

            node_id = self._get_new_node_id()

            async with self._nodes_lock:
                self._nodes[node_id] = created_nodes[node_id] = Node(
                    node_id=node_id,
                    state=NodeState.running,
                    tags=tags,
                    internal_ip=ip,
                    ssh_proxy_command=ssh_proxy_command,
                    activity=activity,
                )

        logger.info(f"Creating {count} nodes done")

        # TODO: Consider running tunnel on every head node
        if not self._is_head_node_to_webserver_tunnel_running():
            await self._start_head_node_to_webserver_tunnel()

        return {
            node_id: node.dict(
                exclude={
                    "activity",
                },
            )
            for node_id, node in created_nodes.items()
        }

    def _get_new_node_id(self) -> NodeId:
        node_id = f"node{self._nodes_id_counter}"
        self._nodes_id_counter += 1
        return node_id

    async def terminate_node(self, node_id: NodeId) -> Dict[NodeId, Dict]:
        logger.info(f"Terminating `{node_id}` node...")

        async with self._get_node_context(node_id) as node:
            node.state = NodeState.stopping

        await node.activity.destroy()

        async with self._get_node_context(node_id) as node:
            del self._nodes[node.node_id]

            logger.info(f"Terminating `{node_id}` node done")

            return {node.node_id: node.dict(exclude={"activity"})}

    async def get_non_terminated_nodes_ids(
        self, tags_to_match: Optional[Dict[str, str]] = None
    ) -> List[NodeId]:
        async with self._nodes_lock:
            if tags_to_match is None:
                return list(self._nodes.keys())

            return [
                node_id
                for node_id, node in self._nodes.items()
                if are_dicts_equal(node.tags, tags_to_match)
            ]

    async def is_node_running(self, node_id: NodeId) -> bool:
        async with self._get_node_context(node_id) as node:
            return node.state == NodeState.running

    async def is_node_terminated(self, node_id: NodeId) -> bool:
        async with self._get_node_context(node_id) as node:
            return node.state not in [NodeState.pending, NodeState.running]

    async def get_node_tags(self, node_id: NodeId) -> Dict:
        async with self._get_node_context(node_id) as node:
            return node.tags

    async def get_node_internal_ip(self, node_id: NodeId) -> str:
        async with self._get_node_context(node_id) as node:
            return node.internal_ip

    async def get_ssh_proxy_command(self, node_id: NodeId) -> str:
        async with self._get_node_context(node_id) as node:
            return node.ssh_proxy_command

    async def set_node_tags(self, node_id: NodeId, tags: Tags) -> None:
        async with self._get_node_context(node_id) as node:
            node.tags.update(tags)

    async def get_or_create_default_ssh_key(self, cluster_name: str) -> str:
        ssh_key_path = self._tmp_path / get_default_ssh_key_name(cluster_name)

        if not ssh_key_path.exists():
            logger.info(f"Creating default ssh key for `{cluster_name}`...")

            ssh_key_path.parent.mkdir(parents=True, exist_ok=True)

            # FIXME: Use cryptography module instead of subprocess
            await run_subprocess_output(
                "ssh-keygen", "-t", "rsa", "-b", "4096", "-N", "", "-f", str(ssh_key_path)
            )

            logger.info(
                f"Creating default ssh key for `{cluster_name}` done with path `{ssh_key_path}`"
            )

        # TODO: async file handling
        with ssh_key_path.open("r") as f:
            return str(f.read())

    @asynccontextmanager
    async def _get_node_context(self, node_id: NodeId) -> Iterator[Node]:
        async with self._nodes_lock:
            node = self._nodes.get(node_id)

            if node is None:
                raise NodeNotFound

            yield node

    async def _get_head_node(self) -> Node:
        async with self._nodes_lock:
            for node in self._nodes.values():
                if node.tags.get(TAG_RAY_NODE_KIND) == NODE_KIND_HEAD:
                    return node

    def _print_ssh_command(
        self, ip: str, ssh_proxy_command: str, ssh_user: str, ssh_private_key_path: Path
    ) -> None:
        logger.debug(
            f"Connect to `{ip}` with:\n"
            f"{get_ssh_command(ip, ssh_proxy_command, ssh_user, ssh_private_key_path)}"
        )

    def _is_head_node_to_webserver_tunnel_running(self) -> bool:
        return self._head_node_to_webserver_tunnel_process is not None

    async def _start_head_node_to_webserver_tunnel(self) -> None:
        logger.info("Starting head node to webserver tunnel...")

        head_node = await self._get_head_node()
        port = self._ray_on_golem_port

        self._head_node_to_webserver_tunnel_process = await run_subprocess(
            "ssh",
            "-N",
            "-R",
            f"*:{port}:127.0.0.1:{port}",
            "-o",
            "StrictHostKeyChecking=no",
            "-o",
            "UserKnownHostsFile=/dev/null",
            "-o",
            f"ProxyCommand={head_node.ssh_proxy_command}",
            "-i",
            str(self._ssh_private_key_path),
            f"{self._ssh_user}@{str(head_node.internal_ip)}",
        )
        self._head_node_to_webserver_tunnel_early_exit_task = asyncio.create_task(
            self._on_head_node_to_webserver_tunnel_early_exit()
        )

        logger.info("Starting head node to webserver tunnel done")

    async def _on_head_node_to_webserver_tunnel_early_exit(self) -> None:
        await self._head_node_to_webserver_tunnel_process.communicate()

        logger.warning(f"Head node to webserver tunnel exited prematurely!")

    async def _stop_head_node_to_webserver_tunnel(self) -> None:
        process = self._head_node_to_webserver_tunnel_process

        if process is None or process.returncode is not None:
            logger.info("Not stopping head node to webserver tunnel, as it's not running")
            return

        logger.info("Stopping head node to webserver tunnel...")

        self._head_node_to_webserver_tunnel_early_exit_task.cancel()
        try:
            await self._head_node_to_webserver_tunnel_early_exit_task
        except asyncio.CancelledError:
            pass

        process.terminate()
        await process.wait()

        self._head_node_to_webserver_tunnel_process = None

        logger.info("Stopping head node to webserver tunnel done")



from ray.autoscaler._private.command_runner import SSHCommandRunner as BaseSshCommandRunner


class SSHCommandRunner(BaseSshCommandRunner):
    def remote_shell_command_str(self):
        ssh_str = f'ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o ProxyCommand="{self.ssh_proxy_command}"'

        if self.ssh_private_key:
            ssh_str += f" -i {self.ssh_private_key}"

        ssh_str += f" {self.ssh_user}@{self.ssh_ip}\n"

        return ssh_str



import asyncio
import base64
import hashlib
import json
import logging
import platform
from collections import defaultdict
from dataclasses import dataclass
from datetime import timedelta
from functools import partial
from pathlib import Path
from typing import AsyncIterator, Dict, List, Optional, Tuple

import aiohttp
import ray
from golem.managers import (
    ActivityManager,
    AddChosenPaymentPlatform,
    AgreementManager,
    BlacklistProviderIdPlugin,
    Buffer,
    DefaultAgreementManager,
    DefaultProposalManager,
    DemandManager,
    LinearAverageCostPricing,
    MapScore,
    NegotiatingPlugin,
    PayAllPaymentManager,
    PaymentManager,
    ProposalManager,
    ProposalManagerPlugin,
    ProposalScorer,
    RefreshingDemandManager,
    RejectIfCostsExceeds,
    ScoringBuffer,
    WorkContext,
)
from golem.managers.proposal.plugins.linear_coeffs import LinearCoeffsCost
from golem.node import SUBNET, GolemNode
from golem.payload import ManifestVmPayload, Payload, constraint, prop
from golem.resources import Activity, Network, ProposalData
from pydantic import BaseModel, Field
from yarl import URL

from ray_on_golem.server.exceptions import RayOnGolemServerError, RegistryRequestError
from ray_on_golem.server.models import DemandConfigData, NodeConfigData
from ray_on_golem.server.services.golem.manifest import get_manifest
from ray_on_golem.server.services.golem.provider_data import PROVIDERS_BLACKLIST, PROVIDERS_SCORED
from ray_on_golem.server.services.utils import get_ssh_command

logger = logging.getLogger(__name__)


class ManagerStack(BaseModel):
    payment_manager: Optional[PaymentManager]
    demand_manager: Optional[DemandManager]
    proposal_manager: Optional[ProposalManager]
    agreement_manager: Optional[AgreementManager]
    activity_manager: Optional[ActivityManager]
    extra_proposal_plugins: List[ProposalManagerPlugin] = Field(default_factory=list)
    extra_proposal_scorers: List[ProposalScorer] = Field(default_factory=list)

    class Config:
        arbitrary_types_allowed = True

    @property
    def _managers(self):
        return [
            self.payment_manager,
            self.demand_manager,
            self.proposal_manager,
            self.agreement_manager,
            self.activity_manager,
        ]

    async def start(self) -> None:
        logger.info("Starting stack managers...")

        for manager in self._managers:
            if manager is not None:
                await manager.start()

        logger.info("Starting stack managers done")

    async def stop(self) -> None:
        logger.info("Stopping stack managers...")

        for manager in reversed(self._managers):
            if manager is not None:
                try:
                    await manager.stop()
                except Exception:
                    logger.exception(f"{manager} stop failed!")

        logger.info("Stopping stack managers done")


class GolemService:
    def __init__(self, websocat_path: Path, registry_stats: bool):
        self._websocat_path = websocat_path
        self._registry_stats = registry_stats

        self._golem: Optional[GolemNode] = None
        self._network: Optional[Network] = None
        self._yagna_appkey: Optional[str] = None
        self._stacks: Dict[str, ManagerStack] = {}
        self._stacks_locks = defaultdict(asyncio.Lock)

    async def init(self, yagna_appkey: str) -> None:
        logger.info("Starting GolemService...")

        self._golem = GolemNode(app_key=yagna_appkey)
        self._yagna_appkey = yagna_appkey
        await self._golem.start()

        self._network = await self._golem.create_network(
            "192.168.0.1/24"
        )  # will be retrieved from provider_config
        await self._golem.add_to_network(self._network)

        logger.info("Starting GolemService done")

    async def shutdown(self) -> None:
        logger.info("Stopping GolemService...")

        await asyncio.gather(
            *[self._remove_stack(stack_hash) for stack_hash in self._stacks.keys()]
        )

        await self._golem.aclose()
        self._golem = None

        logger.info("Stopping GolemService done")

    # TODO: Remove stack when last activity was terminated instead of relying on self shutdown
    async def _remove_stack(self, stack_hash: str) -> None:
        async with self._stacks_locks[stack_hash]:
            # FIXME make sure to terminate running activities
            await self._stacks[stack_hash].stop()

            del self._stacks[stack_hash]
            del self._stacks_locks[stack_hash]

    async def _get_or_create_stack_from_node_config(
        self, node_config: NodeConfigData, budget: float, network: str
    ) -> ManagerStack:
        stack_hash = self._get_hash_from_node_config(node_config)

        async with self._stacks_locks[stack_hash]:
            stack = self._stacks.get(stack_hash)
            if stack is None:
                logger.info(f"Creating new stack `{stack_hash}`...")

                self._stacks[stack_hash] = stack = await self._create_stack(
                    node_config, budget, network
                )
                await stack.start()

                logger.info(f"Creating new stack `{stack_hash}` done")

            return stack

    @staticmethod
    def _get_hash_from_node_config(node_config: NodeConfigData) -> str:
        return hashlib.md5(node_config.json().encode()).hexdigest()

    async def _create_stack(
        self, node_config: NodeConfigData, budget: float, network: str
    ) -> ManagerStack:
        stack = ManagerStack()

        payload = await self._get_payload_from_demand_config(node_config.demand)

        self._apply_cost_management_avg_usage(stack, node_config)
        self._apply_cost_management_hard_limits(stack, node_config)

        stack.payment_manager = PayAllPaymentManager(self._golem, budget=budget, network=network)
        stack.demand_manager = RefreshingDemandManager(
            self._golem,
            stack.payment_manager.get_allocation,
            payload,
            demand_expiration_timeout=timedelta(hours=8),
        )
        stack.proposal_manager = DefaultProposalManager(
            self._golem,
            stack.demand_manager.get_initial_proposal,
            plugins=(
                BlacklistProviderIdPlugin(PROVIDERS_BLACKLIST.get(network, set())),
                *stack.extra_proposal_plugins,
                ScoringBuffer(
                    min_size=50,
                    max_size=1000,
                    fill_at_start=True,
                    proposal_scorers=(
                        *stack.extra_proposal_scorers,
                        MapScore(partial(self._score_with_provider_data, network=network)),
                    ),
                    update_interval=timedelta(seconds=10),
                ),
                NegotiatingPlugin(
                    proposal_negotiators=(AddChosenPaymentPlatform(),),
                ),
                Buffer(min_size=0, max_size=4, fill_concurrency_size=4),
            ),
        )
        stack.agreement_manager = DefaultAgreementManager(
            self._golem, stack.proposal_manager.get_draft_proposal
        )

        return stack

    async def _get_payload_from_demand_config(self, demand_config: DemandConfigData) -> Payload:
        @dataclass
        class CustomManifestVmPayload(ManifestVmPayload):
            subnet_constraint: str = constraint("golem.node.debug.subnet", "=", default=SUBNET)
            debit_notes_accept_timeout: int = prop(
                "golem.com.payment.debit-notes.accept-timeout?", default=240
            )

        image_url, image_hash = await self._get_image_url_and_hash(demand_config)

        manifest = get_manifest(image_url, image_hash)
        manifest = base64.b64encode(json.dumps(manifest).encode("utf-8")).decode("utf-8")

        params = demand_config.dict(exclude={"image_hash", "image_tag"})
        params["manifest"] = manifest

        payload = CustomManifestVmPayload(**params)

        return payload

    async def _get_image_url_and_hash(self, demand_config: DemandConfigData) -> Tuple[URL, str]:
        image_tag = demand_config.image_tag
        image_hash = demand_config.image_hash

        if image_tag is not None and image_hash is not None:
            raise RayOnGolemServerError(
                "Only one of `image_tag` and `image_hash` parameter should be defined!"
            )

        if image_hash is not None:
            image_url = await self._get_image_url_from_hash(image_hash)
            return image_url, image_hash

        if image_tag is None:
            python_version = platform.python_version()
            ray_version = ray.__version__
            image_tag = f"golem/ray-on-golem:py{python_version}-ray{ray_version}"

        return await self._get_image_url_and_hash_from_tag(image_tag)

    async def _get_image_url_from_hash(self, image_hash: str) -> URL:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                f"https://registry.golem.network/v1/image/info",
                params={"hash": image_hash, "count": str(self._registry_stats).lower()},
            ) as response:
                response_data = await response.json()

                if response.status == 200:
                    return URL(response_data["http"])
                elif response.status == 404:
                    raise RegistryRequestError(f"Image hash `{image_hash}` does not exist")
                else:
                    raise RegistryRequestError("Can't access Golem Registry for image lookup!")

    async def _get_image_url_and_hash_from_tag(self, image_tag: str) -> Tuple[URL, str]:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                f"https://registry.golem.network/v1/image/info",
                params={"tag": image_tag, "count": str(self._registry_stats).lower()},
            ) as response:
                response_data = await response.json()

                if response.status == 200:
                    return response_data["http"], response_data["sha3"]
                elif response.status == 404:
                    raise RegistryRequestError(f"Image tag `{image_tag}` does not exist")
                else:
                    raise RegistryRequestError("Can't access Golem Registry for image lookup!")

    def _apply_cost_management_avg_usage(
        self, stack: ManagerStack, node_config: NodeConfigData
    ) -> None:
        cost_management = node_config.cost_management

        if cost_management is None or not cost_management.is_average_usage_cost_enabled():
            logger.debug("Cost management based on average usage is not enabled")
            return

        linear_average_cost = LinearAverageCostPricing(
            average_cpu_load=node_config.cost_management.average_cpu_load,
            average_duration=timedelta(
                minutes=node_config.cost_management.average_duration_minutes
            ),
        )

        stack.extra_proposal_scorers.append(
            MapScore(linear_average_cost, normalize=True, normalize_flip=True),
        )

        max_average_usage_cost = node_config.cost_management.max_average_usage_cost
        if max_average_usage_cost is not None:
            stack.extra_proposal_plugins.append(
                RejectIfCostsExceeds(max_average_usage_cost, linear_average_cost),
            )
            logger.debug("Cost management based on average usage applied with max limits")
        else:
            logger.debug("Cost management based on average usage applied without max limits")

    def _apply_cost_management_hard_limits(
        self, stack: ManagerStack, node_config: NodeConfigData
    ) -> None:
        # TODO: Consider creating RejectIfCostsExceeds variant for multiple values
        proposal_plugins = []
        field_names = {
            "max_initial_price": "price_initial",
            "max_cpu_sec_price": "price_cpu_sec",
            "max_duration_sec_price": "price_duration_sec",
        }

        for cost_field_name, coef_field_name in field_names.items():
            cost_max_value = getattr(node_config.cost_management, cost_field_name, None)
            if cost_max_value is not None:
                proposal_plugins.append(
                    RejectIfCostsExceeds(cost_max_value, LinearCoeffsCost(coef_field_name)),
                )

        if proposal_plugins:
            stack.extra_proposal_plugins.extend(proposal_plugins)
            logger.debug("Cost management based on max limits applied")
        else:
            logger.debug("Cost management based on max limits is not enabled")

    def _score_with_provider_data(
        self, proposal_data: ProposalData, network: str
    ) -> Optional[float]:
        provider_id = proposal_data.issuer_id

        try:
            prescored_providers = PROVIDERS_SCORED[network]
            provider_pos = prescored_providers.index(provider_id)
        except (KeyError, ValueError):
            return 0

        # Gives pre-scored providers from 0.5 to 1.0 score
        return 0.5 + (0.5 * (provider_pos / len(prescored_providers)))

    async def _start_activity(self, context: WorkContext, ip: str):
        logger.info(f"Deploying image on `{context._activity}`")
        deploy_args = {"net": [self._network.deploy_args(ip)]}
        await context.deploy(deploy_args, timeout=timedelta(minutes=5))

        logger.info(f"Starting `{context._activity}`")
        await context.start()

    async def _upload_node_configuration(
        self, context: WorkContext, ip: str, ssh_public_key_data: str
    ):
        logger.info(f"Running initial commands on `{context._activity}`")
        hostname = ip.replace(".", "-")
        await context.run("echo 'ON_GOLEM_NETWORK=1' >> /etc/environment")
        await context.run(f"echo 'NODE_IP={ip}' >> /etc/environment")
        await context.run(f"hostname '{hostname}'")
        await context.run(f"echo '{hostname}' > /etc/hostname")
        await context.run(f"echo '{ip} {hostname}' >> /etc/hosts")
        await context.run("mkdir -p /root/.ssh")
        await context.run(f'echo "{ssh_public_key_data}" >> /root/.ssh/authorized_keys')

    async def _start_ssh_server(self, context: WorkContext):
        logger.info(f"Starting ssh service on `{context._activity}`")
        await context.run("service ssh start")

    async def _verify_ssh_connection(
        self,
        activity: Activity,
        ip: str,
        ssh_proxy_command: str,
        ssh_user: str,
        ssh_private_key_path: Path,
    ) -> None:
        ssh_command = get_ssh_command(ip, ssh_proxy_command, ssh_user, ssh_private_key_path)

        logger.debug(f"Verifying ssh connection with:\n{ssh_command} uptime")

        process = await asyncio.create_subprocess_shell(
            f"{ssh_command} uptime",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        stdout, stderr = await process.communicate()

        msg = f"Validation of ssh connection done on {activity}:\n[cmd exited with {process.returncode}]"
        if stdout:
            msg += f" [stdout] {stdout.decode().strip()}"
        if stderr:
            msg += f" [stderr] {stderr.decode().strip()}"

        if process.returncode != 0:
            logger.warning(msg)
            raise Exception(msg)

        logger.info(msg)

    async def create_activities(
        self,
        node_config: NodeConfigData,
        count: int,
        ssh_public_key: str,
        ssh_user: str,
        ssh_private_key_path: Path,
        budget: float,
        network: str,
    ) -> AsyncIterator[Tuple[Activity, str, str]]:
        stack = await self._get_or_create_stack_from_node_config(node_config, budget, network)

        coros = [
            self._create_activity_retry(stack, ssh_public_key, ssh_user, ssh_private_key_path)
            for _ in range(count)
        ]

        for coro in asyncio.as_completed(coros):
            result = await coro
            await self._network.refresh_nodes()

            yield result

    async def _create_activity_retry(
        self,
        stack: ManagerStack,
        public_ssh_key: str,
        ssh_user: str,
        ssh_private_key_path: Path,
    ) -> Tuple[Activity, str, str]:
        while True:
            try:
                return await self._create_activity(
                    stack, public_ssh_key, ssh_user, ssh_private_key_path
                )
            except Exception:
                logger.warning("Failed to create activity, retrying", exc_info=True)

    async def _create_activity(
        self,
        stack: ManagerStack,
        public_ssh_key: str,
        ssh_user: str,
        ssh_private_key_path: Path,
    ) -> Tuple[Activity, str, str]:
        logger.info(f"Creating new activity...")

        agreement = await stack.agreement_manager.get_agreement()
        activity = await agreement.create_activity()
        try:
            ip = await self._network.create_node(activity.parent.parent.data.issuer_id)
            connection_uri = self._get_connection_uri(ip)
            ssh_proxy_command = self._get_ssh_proxy_command(connection_uri)

            work_context = WorkContext(activity)
            await self._start_activity(work_context, ip)
            await self._upload_node_configuration(work_context, ip, public_ssh_key)
            await self._start_ssh_server(work_context)

            await self._verify_ssh_connection(
                activity, ip, ssh_proxy_command, ssh_user, ssh_private_key_path
            )
        except Exception:
            await activity.destroy()
            raise

        logger.info(f"Creating new activity done with `{activity}` on ip `{ip}`")

        return activity, ip, ssh_proxy_command

    def _get_connection_uri(self, ip: str) -> URL:
        network_url = URL(self._network.node._api_config.net_url)
        return network_url.with_scheme("ws") / "net" / self._network.id / "tcp" / ip

    def _get_ssh_proxy_command(self, connection_uri: URL) -> str:
        # Using single quotes for JWT token as double quotes are causing problems with CLI character escaping in ray
        return f"{self._websocat_path} asyncstdio: {connection_uri}/22 --binary -H=Authorization:'Bearer {self._yagna_appkey}'"
